//==============================================================================
// wgs-simulator.cc
// ----------------------------------------------------------------------------
// Simulate shotgun sequencing of a clone. 
//
//-----------------------------------------------------------------------------
//
//==============================================================================
//.......1.........2.........3.........4.........5.........6.........7.........8

#include "CBT.hh"
#include "RandomNumberGenerator.hh"
#include "WgsSimulator.hh"

//#include "LocalSTL.hh"
//#include "GetOption.hh"
//#include "Sequence.hh"
//#include "FASTAFile.hh"
//#include "RandomNumberGenerator.hh"
//#include "WgsSimulator.hh"

#define SEQ_BUFFER_SIZE  50000

char gbuffer[SEQ_BUFFER_SIZE];


string OptQualityFiles="";
string OptOutFileName = "";
string OptSourceFileName="";
string OptDescriptionFileName="";
int    OptRandomSeed = 0;
int    OptMinP20 = 150;
float  OptErrRate = -1;

//---------------------------------------------------
// * PrintHelp
//---------------------------------------------------
//
void PrintHelp()
{
  printf("\n"
"\n  "
"\n   wgs-simulator.  Requires two text files as input.  The DescriptionFile specifies the "
"\n   number of clones of each type, the size of such clones, and a clone prefix.  One kind of "
"\n   clone is specified per line.  For example:"
"\n"
"\n   2k    1500000 2000 100 1  "
"\n"
"\n   specifies a clone with prefix '2k', that there will be 1.5 million of these clones, that "
"\n   the size of these clones is 2000 bp, with a size StDev of 100, and that paired reads will "
"\n   be generated for each of these clones. "
"\n"
"\n   The second text file required by wgs-simulator is a file of file names listing multiple "
"\n   quality files for real reads (gzipped).  These quality files will be used as templates to "
"\n   determine read size and quality for simulated reads.   "
"\n"
"\n  -s  SourceName         Name of source sequence file. "
"\n  -d  DescriptionFile    Name of description file which lists sizes and numbers of each clone. "
"\n  -o  OutputFile         Name of sequence output.  Quality will be OutputFile.qual"
"\n"
"\n  -q  QualityTemplates   File of file names listing gzipped multiple quality files to use as "
"\n                         a template for the read qualities. "
"\n"
"\n  -e  ErrorRate          Fixed error rate.  Overrides -q if given. "
"\n"
"\n  -P  MinPhred20Count    Minimum phred20 count for any read.   If a read is generated with "
"\n                         P20 less than this value, it will throw it away and generate a new one. "
"\n"
"\n   -r RandomSeed         Seed for random number generator. "
"\n"
"\n    Written By:   K. James Durbin   kdurbin@bcm.tmc.edu"
"\n"
"\n    Version 2.22  Added check that source file exists. "
"\n    Version 2.2   Added InDel support back in.  "
"\n"
"\n    Version 2.11  Fixed bug in read pairs (wasn't doing reverse complement right).  "
"\n    Version 2.1   Changed rand() to random() to get random generators with adequate period. "
"\n    Version 2.0   Rewrite to correct paired end bugs and add real quality files templates. "
"\n                  Some Version 1.0 features temporarialy removed. "
"\n    Version 1.0   Improvements by Harley."
"\n    Version 0.5   Original version, written by James "
"\n");      

  cout<<"LONG_MAX="<<LONG_MAX<<endl;
  cout<<"RAND_MAX="<<RAND_MAX<<endl;
}

//---------------------------------------------------
// * SetupOptions
//---------------------------------------------------
//
void SetupOptions(int argc, char**argv)
{
  // Create parser object and define valid options
  // Options with a : require a parameter. 
  GetOption Parser(argc,argv,"o:s:d:q:r:P:e:h");

  // Handle the options...
  char theOption;
  while((theOption = Parser.NextOption()) != OPTION_EOF){
    switch(theOption){
		case 'e': OptErrRate = Parser.GetFloatArg();break;
		case 'P': OptMinP20 = Parser.GetIntArg();break;
    case 'o': OptOutFileName     = Parser.NextArg(); break;
    case 's': OptSourceFileName = Parser.NextArg(); break;
    case 'd': OptDescriptionFileName = Parser.NextArg(); break;
	 case 'q': OptQualityFiles = Parser.NextArg();break;
	 case 'r': OptRandomSeed = Parser.GetIntArg();break;  
	 case 'h': PrintHelp();exit(1);
    default:
      printf("Error with option '%c'\n",theOption);
      PrintHelp(); exit(1); 
      break;
    }
  }
}


void ParseLine(string &line,string &prefix, int &NumClones,float &MeanCloneSize,float &StDevCloneSize,
					bool &bPaired)
{
  string temp;
  int offset = 0;
  temp = GetNextWSDelimitedString(line,offset,offset);
  prefix = temp;

  temp = GetNextWSDelimitedString(line,offset,offset);
  sscanf(temp.c_str(),"%d",&NumClones);

  temp = GetNextWSDelimitedString(line,offset,offset);
  sscanf(temp.c_str(),"%f",&MeanCloneSize);

  temp = GetNextWSDelimitedString(line,offset,offset);
  sscanf(temp.c_str(),"%f",&StDevCloneSize);

  temp = GetNextWSDelimitedString(line,offset,offset);
  if (temp[0] == '1') bPaired = true;
  else bPaired = false;

}

//---------------------------------------------------
// * main
//---------------------------------------------------
//
int main(int argc, char**argv)
{

  SetupOptions(argc,argv);

  if (!FileExists(OptSourceFileName)){
    cerr<<OptSourceFileName<<" does not exist. Terminating"<<endl;
    exit(1);
  }

  if (!FileExists(OptDescriptionFileName)){
    cerr<<OptDescriptionFileName<<" does not exist. Terminating."<<endl;
    exit(1);
  }

  // RaNdoM
  if (OptRandomSeed==0) {
    // make a number based on the time.
    // Taken from RandomNumberGenerator.cc
    local_time_type seed;
    time(&seed);
    double fs = (double)seed/(double)10000.0;
    long   ifs = (int) fs;
    fs = fs - (double)ifs;
    ifs = (long)(fs*10000);
    OptRandomSeed=ifs;
  }
  cout<<"(RND Seed "<<(unsigned int)OptRandomSeed<<") "<<flush;
  srandom(OptRandomSeed);

  WgsSimulator Sim;
  
  cout<<"Reading source "<<OptSourceFileName<<"..."<<flush;
  ifstream sin;
  gbuffer[0] = '1';



  sin.open(OptSourceFileName.c_str());
  Sim.mSourceSequences.Read(sin,gbuffer,SEQ_BUFFER_SIZE);
  for(int i = 0;i< Sim.mSourceSequences.size();i++){	
	 cout<<Sim.mSourceSequences[i].mName<<"\tsize="<<Sim.mSourceSequences[i].mBaseSize<<endl;
	 Sim.mTotalSeqSize += Sim.mSourceSequences[i].mBaseSize;
  }


	if (OptErrRate >0){
		Sim.mErrRate = OptErrRate;
	}else{
		cout<<"Read trace quality names..."<<flush;
		Sim.mTqTemplate.ReadNames(OptQualityFiles);
		cout<<"done.  NumQualityFiles = "<<Sim.mTqTemplate.mTqFileList.size()<<endl;
	}

  ofstream readFile,qualFile;

  readFile.open(OptOutFileName.c_str());
  if (readFile.fail()){
	 cout<<"ERROR: Unable to open "<<OptOutFileName<<endl;
	 exit(1);
  }

  string QualOutName = OptOutFileName + ".qual";
  qualFile.open(QualOutName.c_str());
  if (qualFile.fail()){
	 cout<<"ERROR: Unable to open "<<QualOutName<<endl;
	 exit(1);
  }

  string ClonePrefix,tempLine;
  int NumClones;
  float MeanCloneSize,StDevCloneSize;
  bool bPaired;

  StartTiming(Sim.mStartTime);

  // Read each line of the description and generate the specified reads. 
  ifstream descFile;

  descFile.open(OptDescriptionFileName.c_str());
  if (descFile.fail()){
	 cout<<"ERROR: Unable to open "<<OptDescriptionFileName<<endl;
	 exit(1);
  }

  do{
	 getline(descFile,tempLine);
	 if (tempLine != ""){
		ParseLine(tempLine,ClonePrefix,NumClones,MeanCloneSize,StDevCloneSize,bPaired);
		cout<<ClonePrefix<<"\t"<<NumClones<<"\t"<<MeanCloneSize<<"\t"<<StDevCloneSize<<"\t"<<bPaired<<endl;
		StartTiming(Sim.mStartTime);

		float saveerr = Sim.mErrRate;
		if (bPaired ){
			Sim.mErrRate = -1;
		}			
		Sim.AddReads(ClonePrefix,NumClones,MeanCloneSize,StDevCloneSize,bPaired,readFile,qualFile);
		Sim.mErrRate = saveerr;
	 }
  }while(!descFile.eof());

  readFile.close();
  qualFile.close();

}






/*
#if 0

  FASTAFile    FAFile;
  WgsSimulator ShotSim;
  Sequence     sourceSeq1;
  Sequence     sourceSeq2;

  SetupOptions(argc,argv);

  cout<<"WGS 1.0\n"<<endl;

  if (OptSource1FileName == "") {
    printf("The source filename is required.\n");
    PrintHelp();
    exit(1);
  }

  // RaNdoM
  if (OptRandomSeed==0) {
    // make a number based on the time.
    // Taken from RandomNumberGenerator.cc
    local_time_type seed;
    time(&seed);
    double fs = (double)seed/(double)10000.0;
    long   ifs = (int) fs;
    fs = fs - (double)ifs;
    ifs = (long)(fs*10000);
    OptRandomSeed=ifs;
    //cout<<"(RND Seed "<<(unsigned int)seed<<") "<<flush;
  }
  srand(OptRandomSeed);


  // setup the palette
  ShotSim.mCxtInsertLength = OptInsertLength;
  ShotSim.mCxtInsertSdev   = OptInsertSdev;
  ShotSim.mCxtPartChimeric = OptPartChimeric;
  ShotSim.mCxtPartU        = OptPartU;
  ShotSim.mCxtPartURError  = OptPartURError;
  ShotSim.mCxtReadAvgLength   = OptReadAvgLength;
  ShotSim.mCxtReadMaxLength   = OptReadMaxLength-1; // Zero based
  ShotSim.mCxtReadSdev     = OptReadSdev;
  ShotSim.mReadChem        = OptReadChem;
  ShotSim.mReadNameCnt     = OptReadNameStart;
  ShotSim.mReadNamePrefix  = OptReadNamePrefix;

  // source seq
  printf("Reading '%s'\n",OptSource1FileName.c_str());
  FAFile.Read(OptSource1FileName,sourceSeq1);
  if (OptSource2FileName!="") {
  printf("Reading '%s'\n",OptSource2FileName.c_str());
    FAFile.Read(OptSource2FileName,sourceSeq2);
  }
  else {
    // all the reads come from the first source
    OptSourceRatio=1.0;
  }

  cout<<"Simulating "<<OptNumReads<<" shotgun reads...\n"<<flush;
  ShotSim.printCxt();
  cout<<"Random Seed = "<<OptRandomSeed<<"\n";

  ShotSim.GenaddCntReads(&sourceSeq1,&sourceSeq2,OptNumReads,OptSourceRatio);

  // Save simulated reads...
  cout<<"Writing seq and qual files...\n"<<flush;
  FAFile.Write(OptOutFileName,ShotSim.mReads);
  OptOutFileName+=".qual";
  FAFile.WriteQual(OptOutFileName,ShotSim.mReadsQual);

  ShotSim.printCounts();

  cout<<"Done!\n";

#endif 

*/

